from pathlib import Path
from typing import Any, Optional, Type, TypeVar

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from accession.database.models import Base

# type hackery for classmethod
_DbSession = TypeVar("_DbSession", bound="DbSession")

ACCESSION_DB_FOLDER = ".accession"
ACCESSION_DB_FILE = "accession.db"


class DbSession:
    """
    Wrapper to encapsulate details of connecting to SQLite via SQLAlchemy
    """

    def __init__(self, uri: str, *args: Any, **kwargs: Any) -> None:
        """
        The engine is already lazy, so no need to wrap it in an @property
        """
        self.engine = create_engine(uri, *args, **kwargs)
        self._session: Optional[Session] = None

    @classmethod
    def with_home_dir_db_path(
        cls: Type[_DbSession], *args: Any, **kwargs: Any
    ) -> _DbSession:
        """
        Creates a new DbSession connected to the DB file in the current user's home
        directory.
        """
        db_path = cls._get_home_dir_db_path()
        db_uri = f"sqlite:///{db_path}"
        return cls(db_uri, *args, **kwargs)

    @classmethod
    def with_in_memory_db(
        cls: Type[_DbSession], *args: Any, **kwargs: Any
    ) -> _DbSession:
        """
        Creates a new DbSession connected to an in-memory, ephermeral DB
        """
        return cls("sqlite:///:memory:", *args, **kwargs)

    @property
    def session(self) -> Session:
        """
        Has the side effect of creating tables that do not already exist.
        """
        if self._session is None:
            Session = sessionmaker(bind=self.engine)
            self._session = Session()
            self._create_tables()
        return self._session

    def _create_tables(self) -> None:
        Base.metadata.create_all(self.engine)

    @staticmethod
    def _get_home_dir_db_path() -> Path:
        """
        Creates a folder in the user's home directory to store the SQLite DB file if it does
        not already exist, and returns the `pathlib.Path` object pointing to where the DB
        file should go.
        """
        accession_db_dir = Path.home() / ACCESSION_DB_FOLDER
        accession_db_dir.mkdir(exist_ok=True)
        return accession_db_dir / ACCESSION_DB_FILE



import logging
import sys

import encode_utils


def logger_factory(
    logger_name: str, log_file_path: str, no_log_file: bool = False
) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.setFormatter(formatter)

    logger.addHandler(stdout_handler)

    eu_debug_logger = logging.getLogger(encode_utils.DEBUG_LOGGER_NAME)
    for hdlr in eu_debug_logger.handlers:
        eu_debug_logger.removeHandler(hdlr)
    eu_debug_logger.addHandler(stdout_handler)

    eu_post_logger = logging.getLogger(encode_utils.POST_LOGGER_NAME)
    for hdlr in eu_post_logger.handlers:
        eu_post_logger.removeHandler(hdlr)
    eu_post_logger.addHandler(stdout_handler)

    if not no_log_file:
        file_handler = logging.FileHandler(log_file_path)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        eu_debug_logger.addHandler(file_handler)
        eu_post_logger.addHandler(file_handler)

    return logger



import json
from pathlib import Path
from unittest.mock import PropertyMock

import pytest
import requests

from accession.analysis import Analysis
from accession.encode_models import EncodeDataset
from accession.metadata import FileMetadata


def mock_queue_qc(qc, *args, **kwargs):
    return qc


@pytest.fixture
def mirna_replicated_analysis(
    mock_accession_gc_backend, mirna_replicated_metadata_path: str
) -> Analysis:  # noqa: F811
    analysis = Analysis(
        FileMetadata(mirna_replicated_metadata_path), backend=mock_accession_gc_backend
    )
    return analysis


@pytest.fixture
def mock_replicated_mirna_accession(mocker, mirna_replicated_analysis, mock_accession):
    """
    Contains a legitimate replicated mirna analysis for purposes of testing mirna qm
    generation without creating any connections to servers.
    """
    mocker.patch.object(mock_accession, "analysis", mirna_replicated_analysis)
    mocker.patch.object(mock_accession, "queue_qc", mock_queue_qc)
    mocker.patch.object(
        mock_accession,
        "dataset",
        new_callable=PropertyMock(
            return_value=EncodeDataset(
                {
                    "@id": "foo",
                    "assay_term_name": "microRNA",
                    "replicates": [
                        {"biological_replicate_number": 1, "status": "released"},
                        {"biological_replicate_number": 2, "status": "in progress"},
                    ],
                }
            )
        ),
    )
    return mock_accession


@pytest.mark.docker
@pytest.mark.filesystem
def test_accession_mirna_replicated(
    accessioner_factory, mirna_replicated_metadata_path, validate_accessioning
):
    factory = accessioner_factory(
        metadata_file=mirna_replicated_metadata_path, assay_name="mirna"
    )
    accessioner, expected_files = next(factory)
    accessioner.accession_steps(force=True)
    validate_accessioning(
        accessioner, expected_files, expected_num_files=12, dataset="ENCSR715NEZ"
    )


@pytest.mark.docker
@pytest.mark.filesystem
def test_accession_mirna_unreplicated(accessioner_factory, validate_accessioning):
    current_dir = Path(__file__).resolve()
    metadata_json_path = (
        current_dir.parent / "data" / "mirna_unreplicated_metadata.json"
    )
    factory = accessioner_factory(metadata_file=metadata_json_path, assay_name="mirna")
    accessioner, expected_files = next(factory)
    accessioner.accession_steps(force=True)
    validate_accessioning(
        accessioner, expected_files, expected_num_files=6, dataset="ENCSR543MWW"
    )


@pytest.fixture
def validate_accessioning(local_encoded_server, api_credentials):
    def _validate_accessioning(
        accessioner, expected_files, expected_num_files, dataset
    ):
        all_files = requests.get(
            f"http://{local_encoded_server}/files/?limit=all&datastore=database",
            headers={"Accept": "application/json"},
            auth=api_credentials,
        ).json()["@graph"]
        all_quality_metrics = [
            accessioner.conn.get(qm["@id"], database=True, frame="object")
            for qm in accessioner.conn.get("/quality-metrics/", database=True)["@graph"]
        ]
        excluded_types = ("reads", "genome index", "genome reference")
        files = [
            f
            for f in all_files
            if f["output_type"] not in excluded_types
            and f["dataset"] == f"/experiments/{dataset}/"
        ]
        assert len(files) == expected_num_files
        shared_keys_to_skip = ["submitted_by", "date_created", "@id", "uuid"]
        file_keys_to_skip = [
            "title",
            "accession",
            "file_size",
            "cloud_metadata",
            "href",
            "s3_uri",
            "no_file_available",
            "analysis_step_version",
            "content_md5sum",
            "matching_md5sum",
            "audit",
            "schema_version",
        ]
        qm_keys_to_skip = ["step_run"]
        for keys_to_skip in (file_keys_to_skip, qm_keys_to_skip):
            keys_to_skip.extend(shared_keys_to_skip)
        for partial_file in files:
            file = accessioner.conn.get(
                partial_file["@id"], frame="embedded", database=True
            )
            quality_metrics = [
                qm
                for qm in all_quality_metrics
                if file["@id"] in qm["quality_metric_of"]
            ]
            aliases = file["aliases"]
            expected = [f for f in expected_files if f["aliases"] == aliases][0]
            for key, expected_value in expected.items():
                if key in file_keys_to_skip:
                    continue
                elif key == "step_run":
                    for step_run_key in ("analysis_step_version", "aliases"):
                        assert file[key][step_run_key] == expected_value[step_run_key]
                    assert file[key]["status"] == "in progress"
                elif key == "status":
                    assert file[key] == "uploading"
                elif key in ("award", "lab"):
                    assert file[key]["@id"] == expected_value["@id"]
                elif key == "file_size":
                    assert file[key] == 3
                elif key == "derived_from":
                    derived_from_aliases = []
                    for i in all_files:
                        if i["@id"] in file[key]:
                            db_file = accessioner.conn.get(i["@id"], database=True)
                            derived_from_aliases.append(
                                (db_file["aliases"], db_file["md5sum"])
                            )
                    expected_aliases = [
                        (f["aliases"], f["md5sum"])
                        for f in expected_files
                        if f["@id"] in expected_value
                    ]
                    assert sorted(derived_from_aliases) == sorted(expected_aliases)
                elif key == "quality_metrics":
                    if not expected_value:
                        assert not quality_metrics
                    else:
                        for expected_qm in expected_value:
                            posted_qms = [
                                qm
                                for qm in quality_metrics
                                if qm["@type"] == expected_qm["@type"]
                            ]
                            assert len(posted_qms) == 1
                            posted_qm = posted_qms[0]
                            for qm_key, qm_value in expected_qm.items():
                                if qm_key == "quality_metric_of":
                                    assert len(qm_value) == len(posted_qm[qm_key])
                                elif qm_key in qm_keys_to_skip:
                                    continue
                                elif qm_key == "status":
                                    assert posted_qm[qm_key] == "in progress"
                                else:
                                    assert qm_value == posted_qm[qm_key]
                else:
                    assert file[key] == expected_value

    return _validate_accessioning


@pytest.mark.filesystem
def test_make_microrna_mapping_qc(mock_replicated_mirna_accession, encode_file_no_qc):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_mapping_qc(
        encode_file_no_qc, gs_file
    )
    assert qc == {"aligned_reads": 5873570}


@pytest.mark.filesystem
def test_make_microrna_correlation_qc_replicated(
    mock_replicated_mirna_accession, encode_file_no_qc
):
    gs_file = [
        i for i in mock_replicated_mirna_accession.analysis.get_files(filekey="tsv")
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_correlation_qc(
        encode_file_no_qc, gs_file
    )
    assert qc == {"Spearman correlation": 0.8885044458946942}


@pytest.mark.filesystem
def test_make_microrna_correlation_qc_unreplicated_returns_none(
    mocker, mock_accession_unreplicated, mirna_replicated_analysis, encode_file_no_qc
):
    mocker.patch.object(
        mock_accession_unreplicated, "analysis", mirna_replicated_analysis
    )
    mocker.patch.object(mock_accession_unreplicated, "queue_qc", mock_queue_qc)
    gs_file = [
        i for i in mock_accession_unreplicated.analysis.get_files(filekey="tsv")
    ][0]
    qc = mock_accession_unreplicated.make_microrna_correlation_qc(
        encode_file_no_qc, gs_file
    )
    assert qc is None


@pytest.mark.filesystem
def test_make_star_qc_metric(mock_replicated_mirna_accession, encode_file_no_qc):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    current_dir = Path(__file__).resolve()
    validation = current_dir.parent / "data" / "validation" / "mirna" / "files.json"
    with open(validation) as f:
        data = json.load(f)
    expected = [i for i in data if i["accession"] == "ENCFF590YAX"][0][
        "quality_metrics"
    ][0]
    qc = mock_replicated_mirna_accession.make_star_qc_metric(encode_file_no_qc, gs_file)
    for k, v in qc.items():
        assert v == expected[k]


@pytest.mark.filesystem
def test_make_microrna_quantification_qc(
    mock_replicated_mirna_accession, encode_file_no_qc
):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_quantification_qc(
        encode_file_no_qc, gs_file
    )
    assert qc == {"expressed_mirnas": 393}



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from caper.caper_args import DEFAULT_CAPER_CONF, get_parser_and_defaults
from caper.caper_labels import CaperLabels
from caper.cromwell_rest_api import CromwellRestAPI, has_wildcard, is_valid_uuid


class CaperHelper:
    """
    Wrapper class around Caper's CromwellRestAPI, since we only need the ability to
    query Cromwell for metadata.
    """

    def __init__(self) -> None:
        self._client: Optional[CromwellRestAPI] = None

    @property
    def client(self) -> CromwellRestAPI:
        if self._client is None:
            _, conf = get_parser_and_defaults(conf_file=DEFAULT_CAPER_CONF)
            self._client = CromwellRestAPI(
                hostname=conf.get("hostname") or CromwellRestAPI.DEFAULT_HOSTNAME,
                port=conf.get("port") or CromwellRestAPI.DEFAULT_PORT,
            )
        return self._client

    def metadata(self, workflow_ids_or_labels: List[str]) -> List[Dict[str, Any]]:
        """
        Don't use the CaperClient here since there is extra stuff in there that we don't
        want or need. The implementation here is copied from
        `caper.caper_client.CaperClient.metadata`, see
        https://github.com/ENCODE-DCC/caper/blob/v1.0.0/caper/caper_client.py
        """
        workflow_ids, labels = self._split_workflow_ids_and_labels(
            workflow_ids_or_labels
        )

        return self.client.get_metadata(workflow_ids, labels, embed_subworkflow=True)

    def _split_workflow_ids_and_labels(
        self, workflow_ids_or_labels: List[str]
    ) -> Tuple[List[str], List[Tuple[str, str]]]:
        workflow_ids: List[str] = []
        labels: List[Tuple[str, str]] = []

        for query in workflow_ids_or_labels:
            if is_valid_uuid(query):
                workflow_ids.append(query)
            else:
                labels.append((CaperLabels.KEY_CAPER_STR_LABEL, query))
                if has_wildcard(query):
                    workflow_ids.append(query)

        return workflow_ids, labels


def caper_conf_exists() -> bool:
    """
    The DEFAULT_CAPER_CONF looks like `~/.caper/default.conf`, need to expand to the
    user's home directory.
    """
    caper_conf = Path(DEFAULT_CAPER_CONF)
    return caper_conf.expanduser().exists()



from accession.metadata_preprocessors import HicMetadataPreprocessor


def test_hic_metadata_preprocessor():
    data = {
        "calls": {
            "ScatterAt87_17": [
                {
                    "subWorkflowMetadata": {
                        "calls": {"hic.align": [{"executionStatus": "Done"}]}
                    }
                }
            ],
            "ScatterAt97_21": [
                {
                    "subWorkflowMetadata": {
                        "calls": {
                            "hic.chimeric_sam_nonspecific": [
                                {"executionStatus": "Done"}
                            ]
                        }
                    }
                }
            ],
            "hic.unnaffected": [{"foo": "bar"}],
        }
    }

    preprocessor = HicMetadataPreprocessor()
    result = preprocessor.process(data)
    assert result == {
        "calls": {
            "hic.align": [
                {
                    "subWorkflowMetadata": {
                        "calls": {"hic.align": [{"executionStatus": "Done"}]}
                    }
                }
            ],
            "hic.chimeric_sam": [
                {
                    "subWorkflowMetadata": {
                        "calls": {
                            "hic.chimeric_sam_nonspecific": [
                                {"executionStatus": "Done"}
                            ]
                        }
                    }
                }
            ],
            "hic.unnaffected": [{"foo": "bar"}],
        }
    }



import argparse
import os
from enum import Enum
from typing import List, Optional, Tuple

from accession import __version__
from accession.accession import accession_factory
from accession.cloud_tasks import QueueInfo
from accession.database.connection import DbSession
from accession.database.query import DbQuery
from accession.metadata import parse_metadata_list
from accession.subcommands.info import info


class Subparsers(Enum):
    Info = "info"


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Accession pipeline outputs, \
                                                 download output metadata for scattering"
    )
    metadata_group = parser.add_mutually_exclusive_group()
    metadata_group.add_argument(
        "-m",
        "--accession-metadata",
        help="path to a metadata json file or a Caper workflow ID/label",
    )
    metadata_group.add_argument(
        "-l",
        "--metadata-list",
        help="A file containing one or more paths to metadata JSON files or Caper workflow IDs/labels, one per line.",
    )
    parser.add_argument(
        "-p",
        "--pipeline-type",
        type=str,
        default=None,
        help="the type of pipeline run being accessioned, e.g. mirna or long_read_rna",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help=(
            "Perform a dry run for accessioning, not posting anything. Has precedence "
            "over -f/--force"
        ),
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force accessioning to post files even if there is md5 duplication",
    )
    parser.add_argument(
        "-s", "--server", default="dev", help="Server the files will be accessioned to"
    )
    parser.add_argument(
        "-a",
        "--annotation-accession",
        help="Accession of annotation to accession to, only valid with `-p segway`",
    )
    parser.add_argument(
        "--no-log-file",
        action="store_true",
        help=(
            "If this flag is specified, don't log to a log file, will still log to "
            "stdout. Will not write to file specified by `--log-file-path`"
        ),
    )
    parser.add_argument(
        "--log-file-path",
        default="accession.log",
        help=(
            "Path to file to write logs to. This will create this file if it does not "
            "exist and append if it already exists."
        ),
    )
    parser.add_argument("--lab", type=str, default=None, help="Lab")
    parser.add_argument("--award", type=str, default=None, help="Award")
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand_name")
    info_parser = subparsers.add_parser(
        name=Subparsers.Info.value,
        help="Obtain accessioning results in a tabular format",
    )
    info_parser.add_argument(
        "-i",
        "--ids",
        help=(
            "One or more experiment accessions or `@id`s, e.g. /experiments/ENCSR123ABC/"
            "or ENCSR456DEF, Cromwell workflow IDs (can specify whole ID or only the "
            "first few digits, a la `git`), `caper-str-label`s, or other Crowmell "
            "workflow labels in key=value form, e.g. my-label=foo. These can be "
            "combined in arbitrary fashion. If no filters are specified, then info "
            "about all accessioning runs will be returned."
        ),
        nargs="+",
    )
    info_parser.add_argument(
        "-d",
        "--date",
        help=(
            "Can be either a single date or a range of dates. Dates should take the "
            "form `MM`, `MM/DD` or `MM/DD/YY`, with zero-padding and year optional. If "
            "the year is left out then the current year will be inserted "
            "automatically. For example, `4/5`, `04/05`, and `4/5/20` are all "
            "equivalent assuming it is 2020. To use a date range, either specify two "
            "valid dates separated by a dash, e.g. `4/5-4/6` or `4-6`, specify one "
            "date with a leading `\\-`, meaning search from the beginning of time to "
            "the provided date, e.g. `\\-4/5/20`, or use a trailing dash to search "
            "from the specifed date to the present."
        ),
    )
    return parser


def check_or_set_lab_award(lab: Optional[str], award: Optional[str]) -> Tuple[str, str]:
    error_msg = (
        "Could not identify {}, make sure you passed in the --{} command line argument"
        "or set the DCC_{} environment variable"
    )
    lab_prop = "lab"
    award_prop = "award"
    if not lab:
        try:
            lab = os.environ[f"DCC_{lab_prop.upper()}"]
        except KeyError as e:
            raise EnvironmentError(
                error_msg.format(lab_prop, lab_prop, lab_prop.upper())
            ) from e
    if not award:
        try:
            award = os.environ[f"DCC_{award_prop.upper()}"]
        except KeyError as e:
            raise EnvironmentError(
                error_msg.format(award_prop, award_prop, award_prop.upper())
            ) from e
    return lab, award


def _check_accession_only_allowed_for_segway(
    accession: Optional[str], pipeline_type: str
) -> None:
    if accession is not None and pipeline_type != "segway":
        raise ValueError("Can only specify accession for pipeline type segway")
    if accession is None and pipeline_type == "segway":
        raise ValueError("Must specify accession for segway pipeline type")


def get_metadatas_from_args(args: argparse.Namespace) -> List[str]:
    """
    If there was only one metadata file/label provided, then returns a list containing
    just that one metadata item, otherwise parses the metadata list file returning the
    list of workflow IDs/labels.
    """
    if args.metadata_list:
        with open(args.metadata_list) as f:
            metadatas = parse_metadata_list(f)
    else:
        metadatas = [args.accession_metadata]
    return metadatas


def main() -> None:
    parser = get_parser()
    args = parser.parse_args()
    lab, award = check_or_set_lab_award(args.lab, args.award)
    _check_accession_only_allowed_for_segway(
        args.annotation_accession, args.pipeline_type
    )
    queue_info = QueueInfo.from_env()
    if args.accession_metadata is not None or args.metadata_list is not None:
        metadatas = get_metadatas_from_args(args)
        for metadata in metadatas:
            accessioner = accession_factory(
                args.pipeline_type,
                metadata,
                args.server,
                lab,
                award,
                log_file_path=args.log_file_path,
                no_log_file=args.no_log_file,
                queue_info=queue_info,
                annotation_accession=args.annotation_accession,
            )
            if accessioner.cloud_tasks_upload_client is not None:
                accessioner.cloud_tasks_upload_client.validate_queue_info()
            accessioner.accession_steps(args.dry_run, force=args.force)
    elif args.subcommand_name == Subparsers.Info.value:
        session = DbSession.with_home_dir_db_path()
        db_query = DbQuery(session)
        info(db_query, args.ids, args.date)


if __name__ == "__main__":
    main()



from unittest.mock import mock_open

import pytest

from accession.accession import AccessionStep, AccessionSteps


@pytest.fixture
def steps(mocker):
    data = """
    {
      "accession.steps": [
        {
          "dcc_step_run": "/analysis-steps/1/",
          "dcc_step_version": "/analysis-step-versions/1-0/",
          "wdl_files": [
            {
              "derived_from_files": [
                {
                  "derived_from_filekey": "fastq",
                  "derived_from_inputs": true,
                  "derived_from_task": "minimap2"
                }
              ],
              "file_format": "bam",
              "filekey": "bam",
              "output_type": "alignments",
              "quality_metrics": [
                  "long_read_rna_mapping"
              ]
            }
          ],
          "wdl_task_name": "minimap2"
        }
      ]
    }
    """
    mocker.patch("builtins.open", mock_open(read_data=data))
    return AccessionSteps("steps.json")


def test_accession_steps_init(steps):
    """
    Once we access steps.content, the _steps is initialized.
    """
    assert not steps._steps
    assert steps.content
    assert steps.content[0].step_run == "/analysis-steps/1/"
    assert steps._steps


def test_accession_steps_content(steps):
    assert isinstance(steps.content, list)
    assert all(isinstance(x, AccessionStep) for x in steps.content)


def test_accession_steps_raw_fastqs_keys_defaults_none(steps):
    assert steps.raw_fastqs_keys is None


def test_file_params(steps):
    file_params = steps.content[0].wdl_files[0]
    assert file_params.filekey == "bam"
    assert file_params.file_format == "bam"
    assert file_params.output_type == "alignments"
    assert file_params.file_format_type is None
    assert isinstance(file_params.derived_from_files, list)
    assert file_params.callbacks == []
    assert file_params.quality_metrics == ["long_read_rna_mapping"]


def test_derived_from_file(steps):
    file_params = steps.content[0].wdl_files[0].derived_from_files[0]
    assert file_params.allow_empty is False
    assert file_params.derived_from_filekey == "fastq"
    assert file_params.derived_from_inputs is True
    assert file_params.derived_from_task == "minimap2"
    assert file_params.derived_from_output_type is None
    assert file_params.disallow_tasks == []



import json
from contextlib import suppress as does_not_raise
from io import StringIO

import pytest

from accession.metadata import (
    CaperMetadata,
    FileMetadata,
    MetadataPreprocessor,
    metadata_factory,
    parse_metadata_list,
)


@pytest.fixture
def file_metadata(mocker, wdl_workflow):
    metadata = {
        "foo": "bar",
        "id": "foo",
        "labels": {"caper-backend": "Local"},
        "submittedFiles": {"workflow": wdl_workflow},
    }
    mocker.patch("builtins.open", mocker.mock_open(read_data=json.dumps(metadata)))
    file_metadata = FileMetadata("foo")
    return file_metadata


def test_metadata_backend_name(file_metadata):
    assert file_metadata.backend_name == "Local"


def test_metadata_get_filtered_labels(mocker):
    mocker.patch.object(
        FileMetadata,
        "original_content",
        new_callable=mocker.PropertyMock(
            return_value={
                "labels": {
                    "my": "label",
                    "caper-str-label": "cool",
                    "cromwell-workflow-id": "123",
                }
            }
        ),
    )
    metadata = FileMetadata("foo.json")
    result = metadata.get_filtered_labels()
    assert result == {"my": "label", "caper-str-label": "cool"}


@pytest.mark.parametrize(
    "prefix,expected", [("", "foo_metadata.json"), ("cool", "cool_foo_metadata.json")]
)
def test_metadata_get_filename(file_metadata, prefix, expected):
    result = file_metadata.get_filename(prefix=prefix)
    assert result == expected


def test_metadata_get_as_attachment(file_metadata):
    result = file_metadata.get_as_attachment(filename_prefix="foo")
    assert isinstance(result.contents, bytes)
    assert result.filename.startswith("foo")
    assert result.mime_type == "application/json"


def test_metadata_get_parsed_workflow(file_metadata):
    result = file_metadata.get_parsed_workflow()
    assert result.workflow.meta["version"] == "v1.2.3"


def test_file_metadata(file_metadata):
    assert file_metadata.original_content["id"] == "foo"
    assert file_metadata.original_content["foo"] == "bar"


@pytest.mark.parametrize(
    "condition,returned,expected",
    [
        (does_not_raise(), [{"foo": "bar"}], {"foo": "bar"}),
        (pytest.raises(ValueError), [], {}),
    ],
)
def test_caper_metadata(mocker, condition, returned, expected):
    caper_metadata = CaperMetadata("foo")
    mocker.patch.object(caper_metadata.caper_helper, "metadata", return_value=returned)
    with condition:
        result = caper_metadata.original_content
        assert result == expected


def test_metadata_with_preprocessors(mocker):
    mocker.patch.object(
        FileMetadata,
        "original_content",
        new_callable=mocker.PropertyMock(return_value={"foo": "bar"}),
    )

    class TestMetadataPreprocessor(MetadataPreprocessor):
        def process(self, metadata):
            return {"foo": metadata["foo"] + "baz"}

    test_preprocessor = TestMetadataPreprocessor()

    metadata = FileMetadata("foo.json", preprocessors=[test_preprocessor])
    result = metadata.content
    assert result == {"foo": "barbaz"}


def test_metadata_factory_file_metadata(tmp_path):
    metadata_file = tmp_path / "metadata.json"
    metadata_file.touch()
    result = metadata_factory(str(metadata_file))
    assert isinstance(result, FileMetadata)


def test_metadata_factory_caper_metadata(mocker):
    """
    Note we need to patch in `accession.metadata` and not `accession.caper_helper`. See
    https://docs.python.org/3/library/unittest.mock.html#where-to-patch
    """
    mocker.patch("accession.metadata.caper_conf_exists", return_value=True)
    result = metadata_factory("foo")
    assert isinstance(result, CaperMetadata)


def test_parse_metadata_list():
    metadata_list = StringIO("   foo   \n\n\tbar")
    result = parse_metadata_list(metadata_list)
    assert result == ["foo", "bar"]


def test_parse_metadata_list_invalid_list_raises():
    metadata_list = StringIO("foo\tbar\n")
    with pytest.raises(ValueError):
        parse_metadata_list(metadata_list)


def test_parse_metadata_list_empty_list_raises():
    metadata_list = StringIO("\t\n\n\n\t\t  \n")
    with pytest.raises(ValueError):
        parse_metadata_list(metadata_list)



from copy import deepcopy
from typing import Any, Dict

from accession.metadata import MetadataPreprocessor


class HicMetadataPreprocessor(MetadataPreprocessor):
    def process(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        new_metadata = deepcopy(metadata)
        for call_name, calls in metadata["calls"].items():
            if "Scatter" in call_name:
                subworkflow_call_keys = calls[0]["subWorkflowMetadata"]["calls"].keys()
                if "hic.align" in subworkflow_call_keys:
                    new_metadata["calls"]["hic.align"] = calls
                    del new_metadata["calls"][call_name]
                for key in (
                    "hic.chimeric_sam_specific",
                    "hic.chimeric_sam_nonspecific",
                ):
                    if key in subworkflow_call_keys:
                        new_metadata["calls"]["hic.chimeric_sam"] = calls
                        del new_metadata["calls"][call_name]
        return new_metadata



import argparse
import os

import pytest

from accession.__main__ import (
    check_or_set_lab_award,
    get_metadatas_from_args,
    get_parser,
)


@pytest.mark.parametrize("lab,award", [("foo", None), (None, "foo")])
def test_check_or_set_lab_award_raises(mocker, lab, award):
    mocker.patch.dict(os.environ, {})
    with pytest.raises(OSError):
        check_or_set_lab_award(lab, award)


@pytest.mark.parametrize(
    "lab,award,environ_prop,expected",
    [
        (None, "bar", "DCC_LAB", ("baz", "bar")),
        ("foo", None, "DCC_AWARD", ("foo", "baz")),
        ("foo", "bar", None, ("foo", "bar")),
    ],
)
def test_check_or_set_lab_award_from_envrion(
    mocker, lab, award, environ_prop, expected
):
    if environ_prop:
        mocker.patch.dict(os.environ, {environ_prop: "baz"})
    result = check_or_set_lab_award(lab, award)
    assert result == expected


def test_get_metadatas_from_args_accession_metadata():
    args = argparse.Namespace(accession_metadata="foo", metadata_list=None)
    result = get_metadatas_from_args(args)
    assert result == ["foo"]


def test_get_metadatas_from_args_metadata_list(mocker):
    mocker.patch("builtins.open")
    mocker.patch("accession.__main__.parse_metadata_list", return_value=["foo", "bar"])
    args = argparse.Namespace(accession_metadata=None, metadata_list="foo")
    result = get_metadatas_from_args(args)
    assert result == ["foo", "bar"]


def test_parser():
    parser = get_parser()
    args = parser.parse_args(["-s", "foo.com", "-m", "meta.json", "-p", "mirna"])
    assert args.server == "foo.com"


def test_parser_info_subcommand_no_args_provided_is_valid():
    parser = get_parser()
    result = parser.parse_args(["info"])
    assert result.ids is None
    assert result.date is None


def test_parser_info_subcommand_date_range_escape():
    parser = get_parser()
    args = parser.parse_args(["info", "-d", "\\-6/24"])
    assert args.ids is None


def test_parser_info_subcommand_multiple_ids():
    parser = get_parser()
    args = parser.parse_args(["info", "-i", "ENCSR123ABC", "21345", "my-wf"])
    assert args.ids == ["ENCSR123ABC", "21345", "my-wf"]


def test_parser_info_subcommand_bare_ids_flag_should_fail():
    parser = get_parser()
    with pytest.raises(SystemExit):
        parser.parse_args(["info", "-i"])



from contextlib import suppress as does_not_raise

import pytest

from accession.accession_steps import FileParams
from accession.encode_models import (
    EncodeAnalysis,
    EncodeAttachment,
    EncodeCommonMetadata,
    EncodeDataset,
    EncodeFile,
    EncodeGenericObject,
    EncodeQualityMetric,
    EncodeStepRun,
    FileStatus,
)


@pytest.fixture
def encode_file():
    return EncodeFile(
        {
            "@id": "1",
            "status": "released",
            "output_type": "reads",
            "dataset": "chip",
            "md5sum": "123",
            "quality_metrics": [{"@type": "Yes"}],
        }
    )


@pytest.fixture
def encode_analysis(encode_common_metadata):
    return EncodeAnalysis(
        files=[EncodeFile({"@id": "/files/1/"}), EncodeFile({"@id": "/files/2/"})],
        documents=[],
        common_metadata=encode_common_metadata,
        workflow_id="123",
    )


@pytest.fixture(scope="module")
def encode_experiment():
    properties = {
        "@id": "/experiments/foo/",
        "assay_term_name": "mirna",
        "replicates": [
            {"biological_replicate_number": 1, "status": "released"},
            {"biological_replicate_number": 3, "status": "released"},
            {"technical_replicate_number": 42, "status": "released"},
        ],
    }
    return EncodeDataset(properties)


@pytest.fixture(scope="module")
def encode_experiment_ignored_replicates():
    properties = {
        "@id": "/experiments/foo/",
        "assay_term_name": "mirna",
        "replicates": [
            {"biological_replicate_number": 1, "status": "released"},
            {"biological_replicate_number": 3, "status": "archived"},
            {"technical_replicate_number": 42, "status": "released"},
            {"technical_replicate_number": 44, "status": "revoked"},
        ],
    }
    return EncodeDataset(properties)


@pytest.fixture
def payload():
    return {"foo": "bar"}


def test_encode_generic_object():
    encode_generic_object = EncodeGenericObject({"@id": "foo"})
    assert encode_generic_object.at_id == "foo"


def test_encode_common_metadata_lab_pi(encode_common_metadata):
    assert encode_common_metadata.lab_pi == "lab"


def test_encode_attachment_encode_attachment_data():
    assert EncodeAttachment.encode_attachment_data(b"foo bar baz") == "Zm9vIGJhciBiYXo="


def test_encode_attachment_get_bytes_from_dict():
    assert EncodeAttachment.get_bytes_from_dict({"a": "b"}) == b'{"a": "b"}'


def test_encode_attachment_encode_attachment_data_string_input():
    with pytest.raises(TypeError):
        EncodeAttachment.encode_attachment_data("foo bar baz")


def test_encode_attachment_make_download_link(encode_attachment):
    assert (
        encode_attachment.make_download_link(additional_extension=".txt")
        == "my_text_file.txt"
    )


def test_encode_attachment_get_portal_object(encode_attachment):
    result = encode_attachment.get_portal_object(
        mime_type="text/plain", additional_extension=".txt"
    )
    assert result == {
        "type": "text/plain",
        "download": "my_text_file.txt",
        "href": "data:text/plain;base64,Zm9vIGJhciBiYXo=",
    }


@pytest.mark.parametrize(
    "other_file,expected",
    [
        (1, False),
        (
            EncodeFile(
                {
                    "@id": "1",
                    "status": "released",
                    "output_type": "reads",
                    "dataset": "chip",
                    "md5sum": "123",
                    "quality_metrics": [{"@type": "Yes"}],
                }
            ),
            True,
        ),
        (
            EncodeFile(
                {
                    "@id": "1",
                    "status": "released",
                    "output_type": "reads",
                    "dataset": "atac",
                    "md5sum": "123",
                    "quality_metrics": [{"@type": "Yes"}],
                }
            ),
            False,
        ),
    ],
)
def test_encode_file_eq(encode_file, other_file, expected):
    result = encode_file == other_file
    assert result == expected


def test_encode_file_properties(encode_file):
    assert encode_file.at_id == "1"
    assert encode_file.status == FileStatus.Released
    assert encode_file.md5sum == "123"
    assert encode_file.output_type == "reads"
    assert encode_file.dataset == "chip"


@pytest.mark.parametrize("key,expected", [("status", "released"), ("name", None)])
def test_encode_file_get(encode_file, key, expected):
    assert encode_file.get(key) == expected


@pytest.mark.parametrize(
    "condition,file,expected",
    [
        (does_not_raise(), EncodeFile({"@id": "1", "step_run": "foo"}), "foo"),
        (does_not_raise(), EncodeFile({"@id": "2", "step_run": {"@id": "bar"}}), "bar"),
        (pytest.raises(ValueError), EncodeFile({"@id": "3"}), ""),
    ],
)
def test_encode_file_step_run_id(condition, file, expected):
    with condition:
        assert file.step_run_id == expected


@pytest.mark.parametrize(
    "condition,new_properties",
    [(does_not_raise(), {"@id": "1"}), (pytest.raises(ValueError), {"@id": "new"})],
)
def test_encode_file_portal_file_setter(encode_file, condition, new_properties):
    with condition:
        encode_file.portal_file = new_properties
        assert encode_file.portal_file == new_properties


@pytest.mark.parametrize("qc_type,expected", [("Yes", True), ("No", False)])
def test_encode_file_has_qc(encode_file, qc_type, expected):
    assert encode_file.has_qc(qc_type) == expected


@pytest.mark.parametrize(
    "files,expected",
    [
        (
            [
                EncodeFile({"@id": 1, "status": "released"}),
                EncodeFile({"@id": 2, "status": "in progress"}),
                EncodeFile({"@id": 3, "status": "deleted"}),
            ],
            2,
        ),
        ([EncodeFile({"@id": 1, "status": "released"})], 1),
    ],
)
def test_encode_file_filter_encode_files_by_status(files, expected):
    assert len(EncodeFile.filter_encode_files_by_status(files)) == expected


def test_encode_file_from_template(encode_common_metadata):
    result = EncodeFile.from_template(
        aliases=["foo"],
        assembly="mm10",
        common_metadata=encode_common_metadata,
        dataset="dataset_id",
        derived_from=["bar"],
        file_params=FileParams(
            {
                "filekey": "foo",
                "file_format": "bam",
                "file_format_type": "bar",
                "output_type": "reads",
                "derived_from_files": [],
            }
        ),
        file_size="123",
        file_md5sum="456",
        step_run_id="step_run_id",
        genome_annotation="V29",
        extras={"extra": "cool"},
        submitted_file_name="baz",
    )
    assert result == {
        "_profile": "file",
        "aliases": ["foo"],
        "assembly": "mm10",
        "award": "award",
        "dataset": "dataset_id",
        "derived_from": ["bar"],
        "extra": "cool",
        "file_format": "bam",
        "file_format_type": "bar",
        "file_size": "123",
        "genome_annotation": "V29",
        "lab": "/labs/lab/",
        "md5sum": "456",
        "output_type": "reads",
        "step_run": "step_run_id",
        "submitted_file_name": "baz",
    }


@pytest.mark.parametrize(
    "other_analysis,expected",
    [
        (
            EncodeAnalysis(
                files=[
                    EncodeFile({"@id": "/files/2/"}),
                    EncodeFile({"@id": "/files/1/"}),
                ],
                documents=[],
                common_metadata=EncodeCommonMetadata(lab="/lab_pies/foo/", award="123"),
                workflow_id="123",
            ),
            True,
        ),
        (
            EncodeAnalysis(
                files=[EncodeFile({"@id": "/files/1/"})],
                documents=[],
                common_metadata=EncodeCommonMetadata(lab="/labs/foo/", award="123"),
                workflow_id="123",
            ),
            False,
        ),
        ("foo", False),
    ],
)
def test_encode_analysis_eq(encode_analysis, other_analysis, expected):
    result = encode_analysis == other_analysis
    assert result == expected


def test_encode_analysis_str(encode_analysis):
    assert str(encode_analysis) == "['/files/1/', '/files/2/']"


def test_encode_analysis_from_files_and_metadata(encode_file, encode_common_metadata):
    result = EncodeAnalysis(
        files=[encode_file],
        common_metadata=encode_common_metadata,
        workflow_id="bar",
        documents=[],
    )
    assert result.aliases == ["lab:bar"]
    assert result.pipeline_version is None


@pytest.mark.parametrize(
    "analysis,expected",
    [
        (
            EncodeAnalysis(
                files=[EncodeFile({"@id": "foo"})],
                documents=[],
                common_metadata=EncodeCommonMetadata(lab="/labs/encode/", award="123"),
                workflow_id="123",
            ),
            {
                "files": ["foo"],
                "_profile": "analysis",
                "aliases": ["encode:123"],
                "documents": [],
                "lab": "/labs/encode/",
                "award": "123",
            },
        ),
        (
            EncodeAnalysis(
                files=[EncodeFile({"@id": "foo"})],
                documents=[EncodeGenericObject({"@id": "bar"})],
                common_metadata=EncodeCommonMetadata(lab="/labs/encode/", award="123"),
                workflow_id="baz",
                pipeline_version="1.0",
            ),
            {
                "_profile": "analysis",
                "aliases": ["encode:baz"],
                "documents": ["bar"],
                "files": ["foo"],
                "pipeline_version": "1.0",
                "lab": "/labs/encode/",
                "award": "123",
            },
        ),
    ],
)
def test_encode_analysis_get_portal_object(analysis, expected):
    result = analysis.get_portal_object()
    assert result == expected


def test_encode_experiment_get_number_of_biological_replicates(encode_experiment):
    assert encode_experiment.get_number_of_biological_replicates() == 2


def test_encode_experiment_get_number_of_biological_replicates_filters_by_status(
    encode_experiment_ignored_replicates
):
    assert (
        encode_experiment_ignored_replicates.get_number_of_biological_replicates() == 1
    )


def test_encode_experiment_get_number_of_technical_replicates(encode_experiment):
    assert encode_experiment.get_number_of_technical_replicates() == 1


def test_encode_experiment_get_number_of_technical_replicates_filters_by_status(
    encode_experiment_ignored_replicates
):
    assert (
        encode_experiment_ignored_replicates.get_number_of_technical_replicates() == 1
    )


def test_encode_experiment_is_replicated(encode_experiment):
    assert encode_experiment.is_replicated is True


def test_encode_experiment_accession(encode_experiment):
    assert encode_experiment.accession == "foo"


def test_encode_experiment_assay_term_name(encode_experiment):
    assert encode_experiment.assay_term_name == "mirna"


def test_encode_experiment_get_patchable_internal_status(encode_experiment):
    result = encode_experiment.get_patchable_internal_status()
    assert result == {
        "internal_status": "pipeline completed",
        "_enc_id": "/experiments/foo/",
    }


def test_encode_experiment_get_patchable_analyses(encode_experiment):
    result = encode_experiment.get_patchable_analyses("foo")
    assert result == {
        "analyses": ["foo"],
        "_enc_id": "/experiments/foo/",
        "_profile": "experiment",
    }


def test_encode_quality_metric_init(payload):
    qm = EncodeQualityMetric(payload, ["my_id"])
    assert qm.files == ["my_id"]
    assert qm.payload == payload


def test_encode_quality_metric_get_portal_object(payload):
    qm = EncodeQualityMetric(payload, ["my_id"])
    result = qm.get_portal_object()
    assert result == {"foo": "bar", "quality_metric_of": ["my_id"]}


def test_encode_step_run_init():
    run = EncodeStepRun({"@id": "1", "name": "foo"})
    assert run.at_id == "1"


def test_encode_step_run_get_portal_object(encode_common_metadata):
    result = EncodeStepRun.get_portal_object(
        aliases=["foo"],
        step_version="/analysis-step-versions/1-0/",
        common_metadata=encode_common_metadata,
    )
    assert result == {
        "_profile": "analysis_step_runs",
        "aliases": ["foo"],
        "analysis_step_version": "/analysis-step-versions/1-0/",
        "award": "award",
        "lab": "/labs/lab/",
    }


def test_encode_document_get_portal_object(mocker, encode_document):
    mocker.patch.object(encode_document.attachment, "get_portal_object")
    result = encode_document.get_portal_object()
    assert result["_profile"] == "document"
    assert result["document_type"] == "workflow metadata"
    assert result["aliases"] == ["encode:foo"]
    assert result["lab"] == "/labs/lab/"
    assert result["award"] == "award"



import datetime

import pytest


def test_get_runs_by_experiment(db_query, run):
    results = db_query.get_runs_by_experiment_at_id("ABC").all()
    assert len(results) == 1


def test_get_runs_by_workflow_id(db_query, run):
    results = db_query.get_runs_by_workflow_id("123").all()
    assert len(results) == 1


@pytest.mark.parametrize("key,value,expected", [("foo", "bar", 1), ("foo", "baz", 0)])
def test_get_runs_by_workflow_label(db_query, run, key, value, expected):
    results = db_query.get_runs_by_workflow_label(key=key, value=value).all()
    assert len(results) == expected


def test_get_runs_by_date_range_one_result_in_range(db_query, run, date):
    start = date - datetime.timedelta(1)
    end = date + datetime.timedelta(1)
    results = db_query.get_runs_by_date_range(start=start, end=end).all()
    assert len(results) == 1


def test_get_runs_by_date_range_no_result_in_range(db_query, run, date):
    start = date - datetime.timedelta(2)
    end = date - datetime.timedelta(1)
    results = db_query.get_runs_by_date_range(start=start, end=end).all()
    assert len(results) == 0


def test_get_runs_by_date_range_one_result_start_no_end(db_query, run, date):
    start = date - datetime.timedelta(1)
    results = db_query.get_runs_by_date_range(start=start).all()
    assert len(results) == 1


def test_get_runs_by_date_range_one_result_end_no_start(db_query, run, date):
    end = date + datetime.timedelta(1)
    results = db_query.get_runs_by_date_range(end=end).all()
    assert len(results) == 1


def test_get_runs_by_date_range_one_result_no_start_no_end(db_query, run, date):
    results = db_query.get_runs_by_date_range().all()
    assert len(results) == 1


def test_get_runs_by_date_range_start_greater_than_end_raises(db_query, run, date):
    start = date - datetime.timedelta(1)
    end = date - datetime.timedelta(2)
    with pytest.raises(ValueError):
        db_query.get_runs_by_date_range(start=start, end=end)


def test_get_all(db_query):
    result = db_query.get_all_runs().all()
    assert len(result) == 1
